/* Copyright (C) 2011 Papavasileiou Dimitris                             
 *                                                                      
 * This program is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation, either version 3 of the License, or    
 * (at your option) any later version.                                  
 *                                                                      
 * This program is distributed in the hope that it will be useful,      
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        
 * GNU General Public License for more details.                         
 *                                                                      
 * You should have received a copy of the GNU General Public License    
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <stdlib.h>
#include <math.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "state.h"
#include "common.h"
#include "config.h"

#define TICKS(n) ((unsigned long)(n) * 15625 / 1000)

#define BUTTONS_MASK ((1 << BUTTONS_A_BIT) |    \
		      (1 << BUTTONS_B_BIT) |	\
		      (1 << BUTTONS_C_BIT) |	\
		      (1 << BUTTONS_D_BIT) |	\
		      (1 << BUTTONS_E_BIT) |	\
		      (1 << BUTTONS_F_BIT))

#define BUTTONS_STATE (~(BUTTONS_PIN | ~BUTTONS_MASK))

static long int ticks, last;
static unsigned char old, clicks, held, button;

static void double_click(void) {
    if (display.tab > 0 && display.tab < 3) {
	if (button == 0 || button == 2) {
	    display.focus += 1;

	    if (display.focus == 1) {
		display.focus += 1;
	    }
	} else if (button == 1) {
	    display.focus -= 1;

	    if (display.focus == 1) {
		display.focus -= 1;
	    }
	}	    

	if (display.focus > 4) {
	    display.focus = 0;
	}

	if (display.focus < 0) {
	    display.focus = 4;
	}
    } else if (display.tab == 3) {
	if (button == 0 || button == 2) {
	    display.focus += 1;
	} else if (button == 1) {
	    display.focus -= 1;
	}

	if (display.focus > 1) {
	    display.focus = 0;
	}	

	if (display.focus < 0) {
	    display.focus = 1;
	}
    } else if (display.tab == 4) {
	if (display.focus == 1 && navigation.editing) {
	    int n = strlen (navigation.waypoint->name);
		
	    if (n < 8 && (button == 0 || button == 3)) {
		strcat (navigation.waypoint->name, "A");
	    }
	} else if (navigation.editing) {
	    if (navigation.editing < 9 && (button == 0 || button == 3)) {
		navigation.editing += 1;
	    }
	} else {
	    if (button == 0 || button == 2) {
		display.focus += 1;
	    } else if (button == 1) {
		display.focus -= 1;
	    }

	    if ((!strcmp (navigation.waypoint->name, ".") &&
		 display.focus > 1) || display.focus > 3) {
		display.focus = 0;
	    }

	    if (display.focus < 0) {
		display.focus = !strcmp (navigation.waypoint->name, ".") ? 1 : 3;
	    }
	}
    } else if (display.tab == 5) {
	if (button == 0 || button == 2) {
	    display.focus += 1;

	    if (statistics.mode == 0 && display.focus == 4) {
		display.focus += 1;
	    }
	} else if (button == 1) {
	    display.focus -= 1;

	    if (statistics.mode == 0 && display.focus == 6) {
		display.focus -= 1;
	    }
	}
	
	if (display.focus > 6) {
	    display.focus = 0;
	}	
	
	if (display.focus < 0) {
	    display.focus = 6;
	}	
    }
}

static void long_click(void) {
    if (display.tab == 1) {
	/* Reset the focused odometer. */
	
	if (display.focus > 1) {
	    reset_odometer(display.focus - 2);
	    distance.running[display.focus - 2] = 0;
	}
    } else if (display.tab == 2) {
	/* Reset the focused timer. */
	
	if (display.focus > 1) {
	    timing.seconds[display.focus - 2] = 0;
	    timing.minutes[display.focus - 2] = 0;
	    timing.ticks[display.focus - 2] = 0;
	    timing.running[display.focus - 2] = 0;
	}
    } else if (display.tab == 4) {
	if (display.focus == 1) {
	    /* If we're not editing start now other wise save
	       the waypoint enter stop editing. */
	    
	    if (!navigation.editing) {
		navigation.editing = 1;
		
		if (!strcmp (navigation.waypoint->name, ".")) {
		    strcpy (navigation.waypoint->name, "A");
		} else {
		    memset (navigation.waypoint->name, 0, 9);
		}
	    } else {
		navigation.editing = 0;

		update_waypoint ();
	    }
	} else if (display.focus == 2 || display.focus == 3) {
	    /* If we're not editing start now other wise save
	       the waypoint enter stop editing. */
	    
	    if (strcmp (navigation.waypoint->name, ".")) {
		if (!navigation.editing) {
		    navigation.editing = 1;
		    navigation.digits = 0;
		} else {
		    navigation.editing = 0;

		    if (display.focus == 2) {
			navigation.waypoint->latitude =
			    navigation.digits * 1e-6;
		    } else {
			navigation.waypoint->longitude =
			    navigation.digits * 1e-6;
		    }

		    update_waypoint ();
		}
	    }
	}
    } else if (display.tab == 5) {
	if (display.focus == 1 && statistics.mode == 2) {
	    statistics.submode = !statistics.submode;
	} else if (display.focus == 2) {
	    statistics.means[statistics.mode].sum = 0;
	    statistics.means[statistics.mode].samples = 0;
	} else if (display.focus == 3) {
	    statistics.peaks[2 * statistics.mode] = 0;
	} else if (display.focus == 4) {
	    statistics.peaks[2 * statistics.mode + 1] = 0;
	}
    }	    
}

static void click(void) {
    /* Buttons 4 through 6 are shortcuts. */
    
    if (!navigation.editing) {
	if (button == 4) {
	    display.tab = 0;
	    display.focus = 0;
	} else if (button == 5) {
	    display.tab = 3;
	    display.focus = 0;
	} else if (button == 6) {
	    display.tab = 5;
	    display.focus = 0;
	}
    }

    /* Any time we're not editing emulate a double click
       when previous or next is pressed. */
    
    if (!navigation.editing && (button == 1 || button == 2)) {
	double_click();
    }
    
    if (display.focus == 0) {
	/* When the tab title is focused use the select or signal
	   button to change the current tab. */
	
	if (button == 0 || button == 3) {
	    display.tab = (display.tab + 1) % 7;
	}
    } else {
	if (display.tab == 1) {
	    /* In the timing tab use the select or signal button to
	       stop/start the timer. */
	    
	    if (button == 0 || button == 3) {
		distance.running[display.focus - 2] =
		    !distance.running[display.focus - 2];
	    }
	} else if (display.tab == 2) {
	    /* In the distance tab use the select or signal button to
	       stop/start the odometer. */
	    
	    if (button == 0 || button == 3) {
		timing.running[display.focus - 2] =
		    !timing.running[display.focus - 2];
	    }
	} else if (display.tab == 3) {
	    /* In the navigation tab use the select or signal button to
	       switch between waypoints. */
	    
	    if (display.focus == 1 && (button == 0 || button == 3)) {
		switch_waypoint();
	    }
	} else if (display.tab == 4) {	    
	    if(display.focus == 1) {
		if (!navigation.editing) {
		    /* If we're not editing use the select or signal
		       button to switch between waypoints. */
		    
		    if (button == 0 || button == 3) {
			switch_waypoint();
		    }
		} else if (button == 3) {
		    /* Simulate a double-click to advance the cursor. */

		    double_click();
		} else if (button < 3) {
		    int n = strlen (navigation.waypoint->name);

		    /* If this is a fresh digit initialize to
		       the character before 'A'. */
		    
		    if (n == 0) {
			navigation.waypoint->name[0] = '@';
			n = 1;
		    }

		    /* Use select or next and previous to browse
		       the alphabet. */
		    
		    if (button == 0 || button == 2) {
			navigation.waypoint->name[n - 1] += 1;
		    } else if (button == 1) {
			navigation.waypoint->name[n - 1] -= 1;
		    }			

		    /* Do some bounds checking. */
		    
		    if (navigation.waypoint->name[n - 1] > 90) {
			navigation.waypoint->name[n - 1] = 65;
		    }

		    if (navigation.waypoint->name[n - 1] < 65) {
			navigation.waypoint->name[n - 1] = 90;
		    }
		}
	    } else if (display.focus == 2 || display.focus == 3) {
		if (navigation.editing) {
		    long power;
		    int i;

		    if (navigation.editing < 9) {
			/* Add the correct power of ten to the coordinate
			   based on the cursor location. */
			
			for (i = 0, power = 1;
			     i < 8 - navigation.editing;
			     i += 1, power *= 10);
		    
			if (button == 0 || button == 2) {
			    navigation.digits += power;			    
			} else if (button == 1) {
			    navigation.digits -= power;
			} else if (button == 3) {
			    /* Simulate a double-click to advance
			      the cursor. */

			    double_click();
			}
			
			/* Do some bounds checking. */
		    
			if (display.focus == 2) {
			    if (navigation.digits >= 90000000) {
				navigation.digits = 90000000;
			    }

			    if (navigation.digits < 0) {
				navigation.digits = 0;
			    }
			}

			if (display.focus == 3) {
			    if (navigation.digits >= 180000000) {
				navigation.digits = 180000000;
			    }

			    if (navigation.digits < 0) {
				navigation.digits = 0;
			    }
			}
		    } else {
			/* Use any of previous,next and signal to switch
			   hemispheres. */
			
			if (button < 3) {
			    navigation.digits *= -1;
			}
		    }
		}
	    }
	} else if (display.tab == 5) {
	    if (button == 0 || button == 3) {
		if (display.focus == 1) {
		    /* Switch mode. */
		    
		    statistics.mode = (statistics.mode + 1) % 3;
		    reset_statistics_graph (1);
		} else if (display.focus == 2) {
		    statistics.hold[statistics.mode] ^= 1;
		} else if (display.focus == 5) {
		    statistics.zooms[0] = (statistics.zooms[0] + 1) % 4;
		    reset_statistics_graph (1);
		} else if (display.focus == 6) {
		    statistics.zooms[1] = (statistics.zooms[1] + 1) % 4;
		    reset_statistics_graph (0);
		}
	    }
	}
    }
}

ISR(INT1_vect)
{
    if (!(TIMSK2 & (1 << OCIE2B))) {
	button = 0;	
	clicks = 0;
	held = 0;
	old = 0;

	ticks = -TCNT2;
	TIMSK2 |= (1 << OCIE2B);
    }

    if (button == 0) {
	last = ticks;
    }
}

ISR(TIMER2_COMPB_vect)
{
    ticks += 255;
}

void check_button_state (void)
{
    char state;

    if (TIMSK2 & (1 << OCIE2B)) {
	if (button == 0) {
	    state = SIGNAL_PIN & (1 << SIGNAL_BIT);
	} else {
	    state = BUTTONS_STATE;
	}

	if ((state != old) && (ticks - last > TICKS(DEBOUNCE_TIME))) {
	    if (state) {
		clicks += 1;
	    }
	    
	    old = state;
	}

	if (state && !held && (ticks - last > TICKS(LONG_CLICK_MIN_TIME))) {
	    held = 1;
	}

	if (state && (ticks - last > TICKS(LONG_CLICK_MAX_TIME))) {
	    held = 0;
	    clicks = 0;
	
	    TIMSK2 &= ~(1 << OCIE2B);
	}

	if (!state && (ticks - last > TICKS(INTERCLICK_TIME))) {
	    TIMSK2 &= ~(1 << OCIE2B);

	    if (held) {
		if (clicks == 1) {
		    long_click();
		}
	    } else {
		if (clicks == 1) {
		    click();
		} else if (clicks == 2) {
		    double_click ();
		}
	    }
	}
    } else if (BUTTONS_STATE > 0) {
	ticks = -TCNT2;
	last = ticks;
	TIMSK2 |= (1 << OCIE2B);

	clicks = 0;
	held = 0;
	old = 0;

	if (BUTTONS_STATE == (1 << BUTTONS_A_BIT)) {
	    button = 1;
	} else if (BUTTONS_STATE == (1 << BUTTONS_B_BIT)) {
	    button = 2;
	} else if (BUTTONS_STATE == (1 << BUTTONS_C_BIT)) {
	    button = 3;
	} else if (BUTTONS_STATE == (1 << BUTTONS_D_BIT)) {
	    button = 4;
	} else if (BUTTONS_STATE == (1 << BUTTONS_E_BIT)) {
	    button = 5;
	} else if (BUTTONS_STATE == (1 << BUTTONS_F_BIT)) {
	    button = 6;
	} else {
	    TIMSK2 &= ~(1 << OCIE2B);
	}
    }
}
