/* Copyright (C) 2011 Papavasileiou Dimitris                             
 *                                                                      
 * This program is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation, either version 3 of the License, or    
 * (at your option) any later version.                                  
 *                                                                      
 * This program is distributed in the hope that it will be useful,      
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        
 * GNU General Public License for more details.                         
 *                                                                      
 * You should have received a copy of the GNU General Public License    
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <avr/io.h>
#include <math.h>

#include "state.h"
#include "common.h"

#define RAD(theta) ((theta) * M_PI / 180)

static char oldminute = -1;

/* Calculate sunlight intensity based on solar azimuth derived
   using a simple approximation by Prof. Richard B. Goldstein
   of Providence College */

static double intensity(double l, double L,
			double M, double d,
			double h, double m){
    double ti, tau, phi, sin_tau, cos_tau;
    double za[] = {-0.5,30.5,58.5,89.5,119.5,150.5,180.5,211.5,242.5,272.5,303.5,333.5};

    /* Convert UTC time provided by the receiver to local time. */
    
    h += L / 15;
    
    if (h < 0) {
	h += 24;
	d -= 1;
    } else if (h > 23) {
	h -= 24;
	d += 1;
    }

    /* Hours +/- offset from noon to degrees from Prime Meridian. */
    
    ti = ((h + m / 60) - 12) * RAD(15);
    
    /* Days since noon March 22nd. */
    
    tau = RAD(360 * (za[(int)M - 1] + 0.5 + d - 82) / 365);

    cos_tau = cos(tau);
    sin_tau = sin(tau);

    /* Sun declination. */
    
    phi = acos(cos_tau * cos_tau + sin_tau * sin_tau * cos(RAD(23.45)));
    
    if (sin_tau < 0) {
	phi = -phi;
    }

    /* altitude = sin^-1(sin theta * sin phi + cos theta * cos phi * cos tau)
       intensity = sin(altitude); */

    return sin(RAD(l)) * sin(phi) + cos(RAD(l)) * cos(phi) * cos(ti);
}

void adjust_brightness(void)
{
    unsigned int day, month, hour, minute;

    /* Get date, time and location from the receiver. */
    
    day = gps.date / 10000;
    month = (gps.date / 100) % 100;
    hour = (long int)gps.time / 10000;
    minute = ((long int)gps.time / 100) % 100;

    if (oldminute != minute && display.brightness == 0) {
	double I;
	    
	I = fmax(intensity(gps.latitude, gps.longitude,
			   month, day, hour, minute), 0);
	
	OCR0A = I > 0 ? I * 255 : 5;

	oldminute = minute;
    } else if (display.brightness > 0) {
	OCR0A = 64 * display.brightness - 1;
    }
}
