/* Copyright (C) 2011 Papavasileiou Dimitris                             
 *                                                                      
 * This program is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation, either version 3 of the License, or    
 * (at your option) any later version.                                  
 *                                                                      
 * This program is distributed in the hope that it will be useful,      
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        
 * GNU General Public License for more details.                         
 *                                                                      
 * You should have received a copy of the GNU General Public License    
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "config.h"
#include "state.h"

static struct {
    double sum;
    int samples;
} window;

static double oldspeed, odometer;
static double v, v_0, theta, theta_0;
static int initialized;

static double sample (void)
{
    double omega, h, V, a_0, a_1;

    h = gps.delta;
    v = gps.speed / 3.6;
    theta = gps.course / 180 * M_PI;

    /* Speed. */
    
    V = 0.5 * (v_0 + v) * 3.6;

    if (V > statistics.peaks[0]) {
	statistics.peaks[0] = V;
    }

    if (!statistics.hold[0]) {
	statistics.means[0].sum += V;
	statistics.means[0].samples += 1;
    }

    /* Don't update acceleration readings
       when the speed is below */
    
    if (gps.speed > ACCELERATION_SAMPLING_THRESHOLD) {
	/* Accelration. */

	a_0 = (v - v_0) / h / 9.80665;

	if (a_0 > statistics.peaks[2]) {
	    statistics.peaks[2] = a_0;
	}

	if (a_0 < statistics.peaks[3]) {
	    statistics.peaks[3] = a_0;
	}

	if (!statistics.hold[1]) {
	    statistics.means[1].sum += a_0;
	    statistics.means[1].samples += 1;
	}

	/* Lateral acceleration. */

	if (theta - theta_0 > M_PI) {
	    omega = (theta - theta_0 - 2 * M_PI) / h;
	} else if (theta - theta_0 < -M_PI) {
	    omega = (theta - theta_0 + 2 * M_PI) / h;
	} else {
	    omega = (theta - theta_0) / h;
	}
	
	a_1 = v * omega / 9.80665;

	if (a_1 > statistics.peaks[4]) {
	    statistics.peaks[4] = a_1;
	}

	if (a_1 < statistics.peaks[5]) {
	    statistics.peaks[5] = a_1;
	}

	if (!statistics.hold[2]) {
	    statistics.means[2].sum += a_1;
	    statistics.means[2].samples += 1;
	}
    }

    v_0 = v;
    theta_0 = theta;
    
    if (statistics.mode == 0) {
	return V;
    } else if (statistics.mode == 1) {
	return a_0;
    } else if (statistics.mode == 2) {
	return a_1;
    } else {
	return 0;
    }
}

void reset_statistics_graph (int reset)
{
    int i;

    if (reset) {
	for (i = 0 ; i < 120 ; i += 1) {
	    statistics.graph.samples[i] = 0;
	}
    
	statistics.graph.count = 0;
	
	window.sum = 0;
	window.samples = 0;

	odometer = 0;
    }
    
    statistics.resolutions[0] = pow (10, statistics.zooms[0] - 2);

    if (statistics.mode == 0) {
	statistics.resolutions[1] = 5 * (statistics.zooms[1] + 1);
    } else if (statistics.mode == 1 || statistics.mode == 2) {
	statistics.resolutions[1] = pow (2, statistics.zooms[1] - 1) / 10;
    }
}

void update_statistics (void)
{
    double n;

    if (!initialized) {
	initialized = 1;
	
	statistics.zooms[0] = 1;
	statistics.zooms[1] = 1;

	reset_statistics_graph(1);
    }
    
    if (gps.delta < 1e-3) {
	return;
    }

    n = sample();

    window.sum += n;
    window.samples += 1;

    odometer += 0.5 * (oldspeed + gps.speed) / 3600.0 * gps.delta;

    if (odometer > statistics.resolutions[0]) {
	statistics.graph.samples[statistics.graph.count % 120] =
	    window.sum / window.samples;
	statistics.graph.count += 1;

	window.sum = 0;
	window.samples = 0;

	odometer -= statistics.resolutions[0];
    }

    oldspeed = gps.speed;
}
