/* Copyright (C) 2011 Papavasileiou Dimitris                             
 *                                                                      
 * This program is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation, either version 3 of the License, or    
 * (at your option) any later version.                                  
 *                                                                      
 * This program is distributed in the hope that it will be useful,      
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        
 * GNU General Public License for more details.                         
 *                                                                      
 * You should have received a copy of the GNU General Public License    
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <math.h>
#include "state.h"
#include "common.h"

static char buffer[83], sentence[83], incoming;
static int length;

int update_gps_data (void)
{
    char *token, *p;
    unsigned char sum = 0x00, check = 0xff;

    /* Enabling the following piece of code will configure the GPS
       receiver when the device starts up.  Disable it again when
       you're done to keep this from happening at every startup.
       In order for this to work the value of USART_BAUDRATE in
       main.c must be the current baud rate setting of the receiver
       (or it won't understand anything).  These commands set the
       baudrate to 38400 (so USART_BAUDRATE must be set accordingly),
       the refresh rate to 5Hz and select the needed NMEA sentences.
       These work for my MMTK chip based GPS receiver.  Other sentences
       may be required to get other receivers to work.
    */
    
#if 0
    {
    	static char done;

    	if (!done) {
	    char *commands[] = {
		"$PMTK314,0,1,0,1,5,0,0,0,0,0,0,0,0,0,0,0,0*2D\r\n",
		"$PMTK251,38400*27\r\n",
		"$PMTK300,200,0,0,0,0*2F\r\n",
	    };
	    
    	    int i, j;

	    for (j = 0 ; j < sizeof (commands) / sizeof (char *) ; j += 1) {
		for (i = 0 ; i < strlen (commands[j]) ; i += 1) {
		    while ((UCSR1A & (1 << UDRE1)) == 0);
		    UDR1 = commands[j][i];
		}
	    }

    	    done = 1;
	}
    }
#endif
    
    if (!incoming) {
	return 0;
    }
    
    p = strchr (sentence, '*');

    if (p) {
	char *q;

	/* Scan the sentence and xor all characters. */
	    
	for (q = sentence, sum = 0;
	     q < p;
	     sum = sum ^ *(unsigned char *)q, q += 1);

	/* Isolate sentence and checksum. */
	    
	*p = '\0';
	*(p + 3) = '\0';

	/* Convert ckecksum into a number and compare. */
	    
	check = (unsigned char)strtoul (p + 1, NULL, 16);
    }

    /* Discard data if DOA. */
    
    if (1 || check == sum) {
	p = sentence;
	token = strsep(&p, ",");    /* Header. */
    
	if (!strcmp (token, "GPRMC")) {
	    token = strsep(&p, ",");    /* Time. */

	    if (gps.time > 0) {
		gps.delta = fmod (atof (token), 100) - fmod (gps.time, 100);
	    
		if (gps.delta < 0) {
		    gps.delta += 60;
		}
	    }
		
	    gps.time = atof (token);

	    token = strsep(&p, ",");    /* Status. */
	    token = strsep(&p, ",");    /* Latitude. */
	    token = strsep(&p, ",");    /* North/South. */
	    token = strsep(&p, ",");    /* Longitude. */
	    token = strsep(&p, ",");    /* East/West. */

	    token = strsep(&p, ",");    /* Speed over ground. */
	    
	    if (gps.speed > 0 || atof (token) > 3) {
	    	gps.speed = atof (token) * 1.852;
	    }

#if 0
	    {
		static int i;

		gps.speed = sin (1e-1 * i) * 120;
		i += 1;
	    }
#endif
	    
	    token = strsep(&p, ",");    /* Course over ground. */
	    gps.course = atof (token);

	    token = strsep(&p, ",");    /* Date. */
	    gps.date = strtoul (token, NULL, 10);

	    token = strsep(&p, ",");    /* Magnetic variation. */
	    token = strsep(&p, ",");    /* Variation sense. */

	    token = strsep(&p, ",");    /* Mode. */

	    update_odometers ();
	    update_statistics ();
	} else if (!strcmp (token, "GPGGA")) {
	    token = strsep(&p, ",");    /* Time. */

	    token = strsep(&p, ",");    /* Latitude. */
	    gps.latitude = atof (token);

	    {
		char deg[4] = "", min[8] = "", *p;

		p = strchr (token, '.');

		if (p) {
		    strncpy (deg, token, p - token - 2);
		    strcpy (min, p - 2);

		    gps.latitude = atof (deg) + atof(min) / 60.0;
		}
	    }

	    token = strsep(&p, ",");    /* North/South. */
	    gps.northsouth = *token;

	    {
		if (toupper(*token) == 'S') {
		    gps.latitude = -gps.latitude;
		}
	    }

	    token = strsep(&p, ",");    /* Longitude. */
	    gps.longitude = atof (token);

	    {
		char deg[4] = "", min[8] = "", *p;

		p = strchr (token, '.');

		if (p) {
		    strncpy (deg, token, p - token - 2);
		    strcpy (min, p - 2);

		    gps.longitude = atof (deg) + atof(min) / 60.0;
		}
	    }
	    
	    token = strsep(&p, ",");    /* East/West. */
	    gps.eastwest = *token;

	    {
		if (toupper(*token) == 'W') {
		    gps.longitude = -gps.longitude;
		}
	    }
	    
	    token = strsep(&p, ",");    /* Fix. */
	    gps.fix = atoi(token);

	    token = strsep(&p, ",");    /* Satellites used. */
	    gps.satellites = (char)atoi (token);

	    token = strsep(&p, ",");    /* HDOP. */

	    token = strsep(&p, ",");    /* Altitude. */
	    gps.altitude = atof (token);

	    token = strsep(&p, ",");    /* Altitude units. */
	    token = strsep(&p, ",");    /* Geoid separation. */
	    token = strsep(&p, ",");    /* Geoid separation units. */
	    token = strsep(&p, ",");    /* Correlation. */
	    token = strsep(&p, ",");    /* Station. */

	    update_navigational_data ();
	} else if (!strcmp (token, "GPGSA")) {
	    token = strsep(&p, ",");    /* Mode 1. */
	    gps.mode[0] = *token;

	    token = strsep(&p, ",");    /* Mode 2. */
	    gps.mode[1] = (char)atoi(token);

	    token = strsep(&p, ",");    /* Id of satellite. */
	    token = strsep(&p, ",");    /* Id of satellite. */
	    token = strsep(&p, ",");    /* Id of satellite. */
	    token = strsep(&p, ",");    /* Id of satellite. */
	    token = strsep(&p, ",");    /* Id of satellite. */
	    token = strsep(&p, ",");    /* Id of satellite. */
	    token = strsep(&p, ",");    /* Id of satellite. */
	    token = strsep(&p, ",");    /* Id of satellite. */
	    token = strsep(&p, ",");    /* Id of satellite. */
	    token = strsep(&p, ",");    /* Id of satellite. */
	    token = strsep(&p, ",");    /* Id of satellite. */
	    token = strsep(&p, ",");    /* Id of satellite. */

	    token = strsep(&p, ",");    /* PDOP. */
	    gps.pdop = atof (token);

	    token = strsep(&p, ",");    /* HDOP. */
	    gps.hdop = atof (token);

	    token = strsep(&p, ",");    /* VDOP. */
	    gps.vdop = atof (token);
	}
    }

    incoming = 0;
    
    return 1;
}

ISR(USART1_RX_vect) 
{
    char letter;

    letter = UDR1;
    
    if (letter == '$' || length > 81) {
	/* This is the beginning of a new sentence
	   Save it if the buffer is free and prepare
	   for the next one. */

	if (!incoming) {
	    memset (sentence, 0, 83);
	    memcpy (sentence, buffer, length);

	    incoming = 1;
	}
	
	memset (buffer, 0, 83);
	length = 0;
    } else {
	buffer[length] = letter;
	length += 1;
    }
}
